package grpc

import (
	"fmt"

	"git.perx.ru/perxis/perxis-go/pkg/errors"
	"git.perx.ru/perxis/perxis-go/proto/common"
	"github.com/hashicorp/go-multierror"
)

func ErrorFromProto(err error, pberr *common.Error) error {
	if pberr == nil {
		return nil
	}

	if err == nil {
		err = errors.New(pberr.Message)
	}

	// Должен быть первым в цепочке
	err = badRequestFromPb(err, pberr)

	if pberr.ErrorCode != 0 {
		err = errors.WithCode(err, pberr.ErrorCode)
	}

	if pberr.ErrorId != "" {
		err = errors.SetID(err, pberr.ErrorId)
	}

	if pberr.Domain != "" {
		err = errors.WithDomain(err, pberr.Domain)
	}

	if pberr.Details != "" {
		err = errors.WithDetail(err, pberr.Details)
	}

	err = helpFromPb(err, pberr)

	var errs []error
	for _, e := range pberr.Errors {
		errs = append(errs, ErrorFromProto(nil, e))
	}
	if len(errs) > 0 {
		err = errors.WithErrors(err, errs...)
	}

	return err
}

func ProtoFromError(err error) *common.Error {
	if err == nil {
		return nil
	}

	err = errors.WithID(err)
	pberr := &common.Error{
		ErrorCode:         errors.GetCode(err),
		ErrorId:           errors.GetID(err),
		Reason:            "", // TBD
		Domain:            errors.GetDomain(err),
		Message:           err.Error(),
		Details:           errors.GetDetail(err),
		Metadata:          nil, // TBD
		BadRequest:        getBadRequest(err),
		DebugInfo:         getDebugInfo(err),
		Help:              getHelp(err),
		LocalizedMessages: nil, // TBD
		Errors:            nil,
	}

	for _, e := range errors.GetErrors(err) {
		pberr.Errors = append(pberr.Errors, ProtoFromError(e))
	}

	return pberr
}

func getHelp(err error) *common.Error_Help {
	if help := errors.GetHelp(err); help != nil {
		pbhelp := &common.Error_Help{}
		for _, hl := range help.Links {
			pbhelp.Links = append(pbhelp.Links, &common.Error_Help_Link{Description: hl.Description, Url: hl.URL})
		}
		return pbhelp
	}
	return nil
}

func helpFromPb(err error, pberr *common.Error) error {
	if pberr.Help == nil || len(pberr.Help.Links) == 0 {
		return err
	}
	h := &errors.Help{}
	for _, l := range pberr.Help.Links {
		h.Links = append(h.Links, errors.HelpLink{Description: l.Description, URL: l.Url})
	}
	return errors.WithHelp(err, h)
}

func getBadRequest(err error) *common.Error_BadRequest {
	br := &common.Error_BadRequest{}

	var merr *multierror.Error
	if errors.As(err, &merr) {
		for _, e := range merr.Errors {
			var errField errors.FieldError
			if errors.As(e, &errField) {
				br.Errors = append(br.Errors, &common.Error_BadRequest_FieldViolation{Field: errField.Field(), Description: errors.Unwrap(errField).Error()})
			}
		}
	} else {
		var errField errors.FieldError
		if errors.As(err, &errField) {
			br.Errors = append(br.Errors, &common.Error_BadRequest_FieldViolation{Field: errField.Field(), Description: errors.Unwrap(errField).Error()})
		}
	}

	if len(br.Errors) > 0 {
		return br
	}

	return nil
}

func getDebugInfo(err error) *common.Error_DebugInfo {
	if trace, ok := errors.GetStackTrace(err); ok {
		di := &common.Error_DebugInfo{}
		for _, t := range trace {
			di.StackTrace = append(di.StackTrace, fmt.Sprintf("%+v", t))
		}
		return di
	}

	return nil
}

func badRequestFromPb(err error, pberr *common.Error) error {
	if pberr.BadRequest == nil || len(pberr.BadRequest.Errors) == 0 {
		return err
	}
	var merr error
	for _, e := range pberr.BadRequest.Errors {
		merr = multierror.Append(merr, errors.WithField(errors.New(e.Description), e.Field))
	}
	return errors.WrapErr(err, merr)
}
